#include <stdio.h>
#include <stdlib.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "type.h"
#include "asm_symbol.h"
#include "asm_code.h"
#include "asm_code_i386.h"

static int asm_code_i386_get_word_size(void)
{
  return 4;
}

static int asm_code_i386_get_pointer_size(void)
{
  return asm_code_i386_get_word_size();
}

static int asm_code_i386_funccall_args_reg_number(void)
{
  return 0;
}

static int asm_code_i386_funccall_args_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_i386_tmp_reg_number(void)
{
  return 3; /* EBX/ESI/EDI (See tmp_regs[]) */
}

static int asm_code_i386_tmp_stack_number(void)
{
  return VALUE_FUNCTION_ARGS_NUM;
}

static int asm_code_i386_function_register_number(void)
{
  return 3; /* EBX/ESI/EDI (See asm_code_i386_function_register_save() */
}

static int asm_code_i386_function_saveparam_number(void)
{
  return 2; /* EBP and return address (See asm_code_i386_function_start() */
}

static int asm_code_i386_stack_align_size(void)
{
  return 16;
}

static int asm_code_i386_stack_correct_size(void)
{
  return 0;
}

static char *r0_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "al";
    case  2: return "ax";
    default: break;
    }
  }
  return "eax";
}

static char *r1_by_size(model_t model)
{
  if (model != NULL) {
    switch (model_get_size(model)) {
    case -1: ERREXIT(1);
    case  1: return "dl";
    case  2: return "dx";
    default: break;
    }
  }
  return "edx";
}

static void asm_code_i386_get_value(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\t$0x%lx, %%eax\n", value);
}

static void asm_code_i386_get_value_r1(FILE *out, long value)
{
  ASM_CODE_OUT(out, "\tmov\t$0x%lx, %%edx\n", value);
}

static void asm_code_i386_get_address_stack(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%esp, %%eax\n");
}

static void asm_code_i386_get_address(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tmov\t$%s, %%eax\n", label);
}

static void asm_code_i386_add_address(FILE *out, int offset)
{
  ASM_CODE_OUT(out, "\tlea\t0x%x(%%eax), %%eax\n", offset);
}

static void asm_code_i386_get_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%eax\n");
}

static void asm_code_i386_set_r1(FILE *out)
{
  ASM_CODE_OUT(out, "\tmov\t%%eax, %%edx\n");
}

static void asm_code_i386_memory_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t0x%x(%%edx), %%%s\n", offset, r0_by_size(model));
}

static void asm_code_i386_memory_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, 0x%x(%%edx)\n", r0_by_size(model), offset);
}

static void asm_code_i386_stack_load(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t0x%x(%%esp), %%%s\n", offset, r0_by_size(model));
}

static void asm_code_i386_stack_store(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, 0x%x(%%esp)\n", r0_by_size(model), offset);
}

static void asm_code_i386_stack_load_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t0x%x(%%esp), %%%s\n", offset, r1_by_size(model));
}

static void asm_code_i386_stack_store_r1(FILE *out, int offset, model_t model)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, 0x%x(%%esp)\n", r1_by_size(model), offset);
}

static void asm_code_i386_stack_expand(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tsub\t$0x%x, %%esp\n", size);
}

static void asm_code_i386_stack_reduce(FILE *out, int size)
{
  ASM_CODE_OUT(out, "\tadd\t$0x%x, %%esp\n", size);
}

static char *funccall_regs[] = {
  "noreg"
};

static void asm_code_i386_funccall_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, %%eax\n", funccall_regs[number]);
}

static void asm_code_i386_funccall_reg_store(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%eax, %%%s\n", funccall_regs[number]);
}

static char *tmp_regs[] = {
  "ebx", "esi", "edi",
};

static void asm_code_i386_tmp_reg_load(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, %%eax\n", tmp_regs[number]);
}

static void asm_code_i386_tmp_reg_save(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%eax, %%%s\n", tmp_regs[number]);
}

static void asm_code_i386_tmp_reg_load_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%%s, %%edx\n", tmp_regs[number]);
}

static void asm_code_i386_tmp_reg_save_r1(FILE *out, int number)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%%s\n", tmp_regs[number]);
}

static int asm_code_i386_sign_extension_char(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovsbl\t%%al, %%eax\n");
  return 0;
}

static int asm_code_i386_sign_extension_uchar(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovzbl\t%%al, %%eax\n");
  return 0;
}

static int asm_code_i386_sign_extension_short(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovswl\t%%ax, %%eax\n");
  return 0;
}

static int asm_code_i386_sign_extension_ushort(FILE *out)
{
  ASM_CODE_OUT(out, "\tmovzwl\t%%ax, %%eax\n");
  return 0;
}

static int asm_code_i386_sign_extension_int(FILE *out)
{
  return -1;
}

static int asm_code_i386_sign_extension_uint(FILE *out)
{
  return -1;
}

static int asm_code_i386_calc_inv(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tnot\t%%eax\n");
  return 0;
}

static int asm_code_i386_calc_minus(FILE *out, model_t model, model_t model_arg0)
{
  ASM_CODE_OUT(out, "\tneg\t%%eax\n");
  return 0;
}

static int asm_code_i386_calc_op1(FILE *out, c_type_t type, model_t model, model_t model_arg0)
{
  return -1;
}

static void asm_code_i386_calc_add(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tadd\t%%edx, %%eax\n");
}

static void asm_code_i386_calc_sub(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tsub\t%%edx, %%eax\n");
}

static void asm_code_i386_calc_and(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tand\t%%edx, %%eax\n");
}

static void asm_code_i386_calc_or(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tor\t%%edx, %%eax\n");
}

static void asm_code_i386_calc_xor(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\txor\t%%edx, %%eax\n");
}

static int asm_code_i386_calc_mul(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\timul\t%%%s\n", r1_by_size(model_arg0));
  return 0;
}

static int asm_code_i386_calc_div(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%ecx\n");
  if (model_arg0->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tmov\t$0, %%edx\n");
  } else {
    ASM_CODE_OUT(out, "\tcltd\n");
  }
  if (model->flags & MODEL_FLAG_UNSIGNED) {
    switch (model_get_size(model_arg1)) {
    case -1: ERREXIT(1);
    case  1: ASM_CODE_OUT(out, "\tdiv\t%%cl\n"); break;
    case  2: ASM_CODE_OUT(out, "\tdiv\t%%cx\n"); break;
    default: ASM_CODE_OUT(out, "\tdiv\t%%ecx\n"); break;
    }
  } else {
    ASM_CODE_OUT(out, "\tidiv\t%%ecx\n");
  }
  return 0;
}

static int asm_code_i386_calc_mod(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%ecx\n");
  if (model_arg0->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tmov\t$0, %%edx\n");
  } else {
    ASM_CODE_OUT(out, "\tcltd\n");
  }
  if (model->flags & MODEL_FLAG_UNSIGNED) {
    ASM_CODE_OUT(out, "\tdiv\t%%ecx\n");
  } else {
    ASM_CODE_OUT(out, "\tidiv\t%%ecx\n");
  }
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%eax\n");
  return 0;
}

static int asm_code_i386_calc_llshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%ecx\n");
  ASM_CODE_OUT(out, "\tshl\t%%cl, %%eax\n");
  return 0;
}

static int asm_code_i386_calc_rashift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%ecx\n");
  ASM_CODE_OUT(out, "\tsar\t%%cl, %%eax\n");
  return 0;
}

static int asm_code_i386_calc_rlshift(FILE *out, model_t model, model_t model_arg0, model_t model_arg1)
{
  ASM_CODE_OUT(out, "\tmov\t%%edx, %%ecx\n");
  ASM_CODE_OUT(out, "\tshr\t%%cl, %%eax\n");
  return 0;
}

static int asm_code_i386_calc_op2(FILE *out, c_type_t type, model_t model, model_t model_arg0, model_t model_arg1)
{
  return -1;
}

static void asm_code_i386_branch(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tjmp\t%s\n", label);
}

static int asm_code_i386_branch_zero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\t$0, %%eax\n");
  ASM_CODE_OUT(out, "\tje\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_nzero(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcmp\t$0, %%eax\n");
  ASM_CODE_OUT(out, "\tjne\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_eq(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tje\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_ne(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjne\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_lt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjl\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_gt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjg\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_le(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjle\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_ge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjge\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_ult(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjb\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_ugt(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tja\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_ule(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjbe\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp_uge(FILE *out, char *label, model_t model)
{
  ASM_CODE_OUT(out, "\tcmp\t%%edx, %%eax\n");
  ASM_CODE_OUT(out, "\tjae\t%s\n", label);
  return 0;
}

static int asm_code_i386_branch_cmp(FILE *out, char *label, c_type_t type, model_t model)
{
  return -1;
}

static int asm_code_i386_function_call(FILE *out, char *label)
{
  ASM_CODE_OUT(out, "\tcall\t%s\n", label);
  return 0;
}

static int asm_code_i386_function_call_set(FILE *out)
{
  /*
   * Standard register(R0) is not used as arguments of the function call,
   * so it can be used as the pointer of the function.
   */
  return -1;
}

static int asm_code_i386_function_call_pointer(FILE *out)
{
  ASM_CODE_OUT(out, "\tcall\t*%%eax\n");
  return 0;
}

static void asm_code_i386_function_start(FILE *out)
{
  ASM_CODE_OUT(out, "\tpush\t%%ebp\n");
  ASM_CODE_OUT(out, "\tmov\t%%esp, %%ebp\n");
}

static void asm_code_i386_function_end(FILE *out)
{
  ASM_CODE_OUT(out, "\tleave\n");
  ASM_CODE_OUT(out, "\tret\n\n");
}

static void asm_code_i386_function_register_save(FILE *out)
{
  ASM_CODE_OUT(out, "\tpush\t%%edi\n");
  ASM_CODE_OUT(out, "\tpush\t%%esi\n");
  ASM_CODE_OUT(out, "\tpush\t%%ebx\n");
}

static void asm_code_i386_function_register_load(FILE *out)
{
  ASM_CODE_OUT(out, "\tpop\t%%ebx\n");
  ASM_CODE_OUT(out, "\tpop\t%%esi\n");
  ASM_CODE_OUT(out, "\tpop\t%%edi\n");
}

struct asm_code_callback asm_code_callback_i386 = {
  asm_code_i386_get_word_size,
  asm_code_i386_get_pointer_size,

  asm_code_i386_funccall_args_reg_number,
  asm_code_i386_funccall_args_stack_number,
  asm_code_i386_tmp_reg_number,
  asm_code_i386_tmp_stack_number,
  asm_code_i386_function_register_number,
  asm_code_i386_function_saveparam_number,

  asm_code_i386_stack_align_size,
  asm_code_i386_stack_correct_size,

  asm_code_i386_get_value,
  asm_code_i386_get_value_r1,
  asm_code_i386_get_address_stack,
  asm_code_i386_get_address,
  asm_code_i386_add_address,
  asm_code_i386_get_r1,
  asm_code_i386_set_r1,

  asm_code_i386_memory_load,
  asm_code_i386_memory_store,
  asm_code_i386_stack_load,
  asm_code_i386_stack_store,
  asm_code_i386_stack_load_r1,
  asm_code_i386_stack_store_r1,
  asm_code_i386_stack_expand,
  asm_code_i386_stack_reduce,

  asm_code_i386_funccall_reg_load,
  asm_code_i386_funccall_reg_store,
  asm_code_i386_tmp_reg_load,
  asm_code_i386_tmp_reg_save,
  asm_code_i386_tmp_reg_load_r1,
  asm_code_i386_tmp_reg_save_r1,

  asm_code_i386_sign_extension_char,
  asm_code_i386_sign_extension_uchar,
  asm_code_i386_sign_extension_short,
  asm_code_i386_sign_extension_ushort,
  asm_code_i386_sign_extension_int,
  asm_code_i386_sign_extension_uint,

  asm_code_i386_calc_inv,
  asm_code_i386_calc_minus,

  asm_code_i386_calc_op1,

  asm_code_i386_calc_add,
  asm_code_i386_calc_sub,
  asm_code_i386_calc_and,
  asm_code_i386_calc_or,
  asm_code_i386_calc_xor,

  asm_code_i386_calc_mul,
  asm_code_i386_calc_div,
  asm_code_i386_calc_mod,
  asm_code_i386_calc_llshift,
  asm_code_i386_calc_rashift,
  asm_code_i386_calc_rlshift,

  asm_code_i386_calc_op2,

  asm_code_i386_branch,
  asm_code_i386_branch_zero,
  asm_code_i386_branch_nzero,

  asm_code_i386_branch_cmp_eq,
  asm_code_i386_branch_cmp_ne,

  asm_code_i386_branch_cmp_lt,
  asm_code_i386_branch_cmp_gt,
  asm_code_i386_branch_cmp_le,
  asm_code_i386_branch_cmp_ge,
  asm_code_i386_branch_cmp_ult,
  asm_code_i386_branch_cmp_ugt,
  asm_code_i386_branch_cmp_ule,
  asm_code_i386_branch_cmp_uge,

  asm_code_i386_branch_cmp,

  asm_code_i386_function_call,
  asm_code_i386_function_call_set,
  asm_code_i386_function_call_pointer,

  asm_code_i386_function_start,
  asm_code_i386_function_end,
  asm_code_i386_function_register_save,
  asm_code_i386_function_register_load,
};
